/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "bits/__generics.h"
#include "vtype.h"

#ifndef LIBCDSB_DICT_H
#define LIBCDSB_DICT_H

typedef int (*dict_access_callback)(vtype_variable key, vtype_variable value, void* data);

/*#####################################################################################################################*/

inline void dict_init(vtype_dict* x) Always_inline__ Nonnull__(1);

inline void dict_init(vtype_dict* x) { x->nodes = (void*)(x->capacity = x->size = 0); }

/*#####################################################################################################################*/

#define dict_pop(x, key, data, callback) libcdsb_dict_find   (x, _LIBCDSB_var(key), data, callback, 1)
#define dict_get(x, key, data, callback) libcdsb_dict_find   (x, _LIBCDSB_var(key), data, callback, 0)
#define dict_update(x, key, value)       libcdsb_dict_update (x, _LIBCDSB_var(key), _LIBCDSB_var(value), 0, 0)
#define dict_inject(x, key, value)       libcdsb_dict_inject (x, _LIBCDSB_var(key), _LIBCDSB_var(value), 0, 0)
#define dict_foreach(x, data, callback)  libcdsb_dict_foreach(x, data, callback, 0)
#define dict_remove(x, key)              dict_pop            (x, key, 0, 0)

#define in_dict(x, key)                 (dict_get(&x, key, 0, 0) == 0)

/*#####################################################################################################################*/

extern bool libcdsb_dict_update       (vtype_dict* x, vtype_variable key, vtype_variable value, void* data, dict_access_callback)             Nonnull__(1);
extern bool libcdsb_dict_inject       (vtype_dict* x, vtype_variable key, vtype_variable value, void* data, dict_access_callback)             Nonnull__(1);
extern int  libcdsb_dict_find         (vtype_dict* x, vtype_variable key,                       void* data, dict_access_callback, bool cut)   Nonnull__(1);
extern int  libcdsb_dict_foreach      (vtype_dict* x,                                           void* data, dict_access_callback, bool flush) Nonnull__(1,3);
extern bool libcdsb_dict_shrink_to_fit(vtype_dict* x)                                                                                         Nonnull__(1);

#endif /* LIBCDSB_DICT_H */
