/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "include.h"

int libcdsb_dict_find(dict_t* x, var_t kvar, void* dt, dict_access_callback callback, bool cut) {
    dnode_t *c, **p;
    int           r;
    var_t      k, v;

    if (x->capacity) {
        c = *(p = x->nodes + (variable_hash(kvar) % x->capacity));

        while (!is_null(c)) {
            k = vnode_peek(&c->key,   c->key_type);
            v = vnode_peek(&c->value, c->value_type);

            if (variable_compare(k, kvar) == 0) {
                r = (callback) ? callback(k, v, dt) : 0;

                if (cut) {
                    *p = c->prev;
                    vnode_free(&c->key,   c->key_type);
                    vnode_free(&c->value, c->value_type);
                    free(c);
                    --x->size;
                }

                return r;
            } else c = *(p = &c->prev);
        }
    }

    return -1;
}


int libcdsb_dict_foreach(dict_t* x, void* dt, dict_access_callback callback, bool flush) {
    dnode_t *c;
    ssize_t  i;
    int      r;
    var_t k, v;

    r = 0;
    i = x->capacity;

    while (i) {
        c = x->nodes[--i];

        while (!is_null(c)) {
            k = vnode_peek(&c->key,   c->key_type);
            v = vnode_peek(&c->value, c->value_type);

            r = callback(k, v, dt);
            c = c->prev;

            if (r) {
                if (!flush) goto end_;
                else goto flush_loop_;
            } else if (flush) {
                vnode_free(&x->nodes[i]->key,   x->nodes[i]->key_type);
                vnode_free(&x->nodes[i]->value, x->nodes[i]->value_type);
                free(x->nodes[i]);

                x->nodes[i] = c;
            }
        }
    }

    if (flush) {
        while (i) {
            --i;

            while (!is_null(x->nodes[i])) { flush_loop_:
                vnode_free(&x->nodes[i]->key,   x->nodes[i]->key_type);
                vnode_free(&x->nodes[i]->value, x->nodes[i]->value_type);

                c = x->nodes[i]->prev;
                free(x->nodes[i]);

                x->nodes[i] = c;
            }
        }

        free(x->nodes);
        memset(x, 0, sizeof(*x));
    }

  end_:
    return r;
}
