/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include "include.h"

bool libcdsb_map_update(map_t* x, vtype_variable key, vtype_variable value, void* dt, map_access_callback callback) {
    int    cmp;
    mnode_t* n;
    mnode_t* p;
    vtype_variable k;

    if (!mnode_is_empty(n = x->root)) {
        do {
            p = n;
            k = vnode_peek(&n->key, x->type);
            cmp = variable_compare(key, k);

            if (cmp == 0) {
                if (callback) callback(k, vnode_peek(&n->value, n->type), dt);

                vnode_free(&n->value, n->type);

                n->value = vnode_create(value);
                n->type  = value.type;

                return true;
            }

            n = (cmp < 0) ? n->left : n->right;
        } while (!mnode_is_empty(n));

        n = mnode_create(nullptr, p, 1);

        if (cmp < 0) p->left  = n;
        else         p->right = n;

        if (!mnode_is_root(p))
            mnode_fixup(&x->root, n);

    } else n = x->root = mnode_create(nullptr, mnode_empty, 0);

    n->key   = vnode_create_ex(x->type, key);
    n->value = vnode_create(value);
    n->type  = value.type;

    return false;
}


bool libcdsb_map_inject(map_t* x, vtype_variable key, vtype_variable value, void* dt, map_access_callback callback) {
    int    cmp;
    mnode_t* n;
    mnode_t* p;
    vnode_t kn;
    vtype_variable k;

    vnode_attach_ex(&kn, x->type, key);
    n   = x->root;
    key = vnode_peek(&kn, x->type);

    if (!mnode_is_empty(n)) {
        do {
            k   = vnode_peek(&n->key,   x->type);
            p   = n;
            cmp = variable_compare(key, k);

            if (cmp == 0) {
                if (callback) callback(k, vnode_peek(&n->value, n->type), dt);

                vnode_free(&n->key,   x->type);
                vnode_free(&n->value, n->type);

                n->key  = kn;
                n->type = value.type;
                vnode_attach(&n->value, value);

                return true;
            }

            n = (cmp < 0) ? n->left : n->right;
        } while (!mnode_is_empty(n));

        n = mnode_create(kn, p, 1);

        if (cmp < 0) p->left  = n;
        else         p->right = n;

        if (!mnode_is_root(p))
            mnode_fixup(&x->root, n);

    } else n = x->root = mnode_create(kn, mnode_empty, 0);

    vnode_attach(&n->value, value);
    n->type = value.type;

    return false;
}
